import subprocess


def find_subclasses(cls):
    """
    Recursively find all subclasses of given class cls.
    """
    subclasses = cls.__subclasses__()
    for subclass in list(subclasses):
        subclasses.extend(find_subclasses(subclass))
    return list(set(subclasses))


class OutputFormatter:
    terminal = False

    def __init__(self, text):
        if type(text) == list:
            self.lines = text
        else:
            self.lines = text.splitlines()

    def __getattr__(self, name):
        cls = self.__class__
        cls = []
        if self.terminal:
            raise Exception('Unable to chain formatter "%s"; "%s" is terminal' % (
                name, cls.__name__.lower()
            ))
        while cls is not OutputFormatter:
            [cls.append(kls) for kls in cls.__bases__]
            if cls:
                cls = cls.pop()
            else:
                raise Exception("I'm lost and failed to find myself somehow.")
        for formatter in find_subclasses(cls):
            if formatter.__name__.lower() == name.lower():
                return formatter(self.lines[:])
        raise KeyError('formatter not found: %s' % name)

    def __str__(self):
        return '\n'.join(self.lines)


class Head(OutputFormatter):
    def __call__(self, n=10, c=None):
        if c:
            lines = []
            left = c
            for line in self.lines:
                if len(line) < left:
                    # clear to add the entire line
                    lines.append(line)
                else:
                    # append only whats remaining to be read
                    lines.append(line[:len(line) - left])
                left -= len(lines[-1])
                if left <= 0:
                    break
            self.lines = lines
        if n:
            self.lines = self.lines[:n]
        return self


class Tail(Head):
    def __call__(self, n=10, c=None):
        head = self.head
        head.lines.reverse()
        head(n, c)
        self.lines = head.lines
        self.lines.reverse()
        return self


if __name__ == '__main__':
    # testing...
    data = '1a\n2b\n3c\n4d\n5e\n6f\n7g\n8h\n9i\n10j\n11k\n'
    print('* first 3 lines')
    outf = OutputFormatter(data)
    print(outf.head(3))
    print('* 3rd line')
    print(Head(data)(3).tail(1))
    print('* last 3 lines')
    print(outf.tail(3))
    print('* 3rd to last lines')
    print(Tail(data)(3).head(1))
    print(outf.less())



from typing import Union, NamedTuple
import re
import urllib.parse



class ByteSize(NamedTuple):
    num_bytes: int
    unit: str
    value: float

    def __str__(self):
        return f'{self.value} {self.unit}'


class Duration(NamedTuple):
    # the duration in total
    ms: int
    # description, broken out by days/hours/minutes as needed
    desc: str

    def __str__(self):
        return self.desc


class ResponseMeta(NamedTuple):
    duration: Duration
    byte_size: ByteSize
    success: bool
    code: int


class Headers:
    """
    Wrapper around HTTP header management forcing lowercase keys and ensuring duplication.
    """
    def __init__(self, headers: dict = None):
        self._headers = {}
        self.update(headers)

    def as_dict(self):
        return {
            name: val
            for name, val in self._headers.items()
        }

    def get(self, key, default=None):
        key_lower = key.lower()
        if key_lower in self._headers:
            return self._headers[key_lower]
        return default

    def add(self, key:str, val:str, update_only:bool = False):
        key_lower = key.lower()
        if update_only and not self.exists(key_lower):
            raise ValueError(f'No header named "key" exists; cannot add')
        self._headers[key_lower] = val

    def exists(self, key:str) -> bool:
        return key.lower in self._headers

    def remove(self, key:str, ignore_missing=True) -> bool:
        """
        Delete a header and return whether not it was actually deleted.
        """
        key_lower = key.lower()
        if key_lower in self._headers:
            del(self._headers[key_lower])
            return True
        elif not ignore_missing:
            raise ValueError(f'No header named "key" exists; cannot remove')
        return False

    def update(self, headers: dict):
        if not headers:
            return
        for key, val in headers.items():
            self.add(key, val)

    def copy(self):
        return Headers(self._headers.copy())

    def items(self):
        return self._headers.items()


class Response(NamedTuple):
    obj: any
    decoded: any
    data: bytes
    meta: ResponseMeta


class Url(NamedTuple):
    scheme: str
    hostname: str
    port: int
    path: str
    query: str
    params: str
    fragment: str

    def __str__(self):
        url = f'{self.scheme}://{self.hostname}'
        if any([
                (self.scheme == 'https' and self.port != 443),
                (self.scheme == 'http' and self.port != 80),
            ]):
            url+= ':' + str(self.port)
        url += self.path
        if self.params:
            url += f';{self.params}'
        if self.query:
            url += f'?{self.query}'
        if self.fragment:
            url += f'#{self.query}'
        return url

    def validate(self):
        if self.scheme.lower() not in ['http', 'https']:
            raise ValueError('Only HTTP and HTTPS are supported protocols.')
        if self.port < 1 or self.port > 65535:
            raise ValueError(f'Invalid API service port: {port}.')

    @staticmethod
    def parse_str(url_str:str):
        """
        Parses a URL into its components. Allows as little information as
        possible (e.g. just a port, just a path), defaulting to
        http://localhost:80/.
        """
        # urlparse just doesn't do it the "right" way...
        # check for protocol and strip it off if found
        parts = {
            'scheme': None,
            'hostname': None,
            'port': None,
            'path': None,
            'params': None,
            'query': None,
            'fragment': None
        }
        # parse out the scheme, if present
        if re.match('^\w+://', url_str):
            (scheme, url_str) = url_str.split('://', 1)
            parts['scheme'] = scheme.lower()
        # check for a path
        if url_str.find('/') == -1:
            hostname = url_str
            url_str = ''
        else:
            (hostname, url_str) = url_str.split('/', 1)
        # do we have a port in the hostname?
        if hostname.find(':') >= 0:
            # chop out the port
           hostname, parts['port'] = hostname.split(':', 1)
           parts['port'] = int(parts['port'])
           if not parts['scheme']:
               if parts['port'] == 80:
                   parts['scheme'] = 'http'
               elif parts['port'] == 443:
                   parts['scheme'] = 'https'
        if not hostname:
            hostname = 'localhost'
        parts['hostname'] = hostname.lower()
        # let urlparse do the rest of the work on the path w/ a fake domain
        parsed = urllib.parse.urlparse('http://localhost/' + url_str)
        parts['path'] = parsed.path
        parts['params'] = parsed.params
        parts['query'] = parsed.query
        parts['fragment'] = parsed.fragment
        # normalize a bit and supply any defaults
        if not parts['scheme']:
           parts['scheme'] = 'https'
        if not parts['port']:
            if parts['scheme'] == 'https':
                parts['port'] = 443
            else:
                parts['port'] = 80
        else:
            parts['port'] = int(parts['port'])
        url = Url(**parts)
        url.validate()
        return url



import decimal
import functools
import re
import subprocess

from . import types


def get_args(my_args:dict = None, args:dict = None, merge: bool = False) -> dict:
    """
    Returns a dict of args based on the initial args given in `my_args`.
    Ignores extra args in `args` unless `merge` is set.
    """
    final_args = my_args.copy()
    if not args:
        return final_args
    for arg, val in args.items():
        if arg in my_args or merge:
            final_args[arg] = val
    return final_args


def pretty_path(path, absolute=False, no_trailing=True):
    if no_trailing:
        path = path.rstrip('/')
    if absolute:
        path = '/' + path
    regex = re.compile(r'/+')
    path = regex.sub('/', path)
    return path


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def get_duration(ms:int) -> types.Duration:
    """
    Describe a time specified in duration of milliseconds, as an absolute value
    (e.g. negatives are treated as positive).
    """
    ms = abs(ms)
    duration = functools.partial(types.Duration, ms=ms)
    dec_p2 = decimal.Decimal('0.00')

    if ms < 1000:
        return duration(desc=f'{ms}ms')
    secs = ms / 1000
    if secs < 60:
        return duration(
            desc=str(decimal.Decimal(secs).quantize(dec_p2)) + 's'
        )
    # don't worry about MS precision as we go up
    secs = int(secs)
    mins = int(secs / 60)
    secs -= mins * 60
    if mins < 60:
        return duration(desc=f'{mins}m, {secs}m')
    hours = int(mins / 60)
    mins -= hours * 60
    if hours < 24:
        return duration(desc=f'{hours}h, {mins}m, {secs}s')
    days = int(hours / 24)
    hours -= days * 24
    return duration(desc=f'{days}d, {hours}h, {mins}m, {secs}s')


def get_byte_size(val, precision:int = 2) -> types.ByteSize:
    """
    Describe, with arbitrary precision, bytes up to the PB range. Uses a base
    of 1000, not 1024.
    """
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    prev_limit = 0
    best_unit = None
    remainder = val
    for i, unit in enumerate(units):
        best_unit = unit
        unit_limit = 10 ** ((i + 1) * 3)
        if remainder < unit_limit:
            break
        prev_limit = unit_limit
    if prev_limit:
        remainder = round(remainder / prev_limit, precision)
    return types.ByteSize(value=remainder, unit=best_unit, num_bytes=val)

def run(cmd, **popen_args):
    """
    Run a shell command and return tuple of (stdout, stderr, retval).
    """
    try:
        proc = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            **popen_args
        )
        stdout, stderr = proc.communicate()
    except Exception as e:
        return None, str(e), 127
    return stdout.decode('utf-8'), stderr.decode('utf-8'), proc.returncode



#!/usr/bin/env python3

# TODO:
#   * 'grep -v' to selectively hide instead of include
#   * '//foo' floating extraction
#      - e.g. double slashes (or whatever the field separator is set to) designates "floating" paths (e.g. "//foo" matches 'foo' anywhere; "foo//bar" matches "bar" anywhere under "/foo")
#   * filter (or include) results based on values (e.g. "-f 'foo/id >= 3'")
#   * verbose debugging to troubleshoot filtering/extraction

import collections
import json
import re
import sys


def usage():
    sys.stdout.write("""usage: jsonx [ARGS] [JSON_FILE]

Pretty prints JSON by default. If a file is given it will be read for JSON
data; otherwise STDIN will be read instead.

ARGUMENTS:

   -p|--pairs            Convert to name=value for easy variable assignment.
   -h|--help             This information.
   -q|--quiet            If a path cannot be extracted/followed quiety ignore it.
   -j|--json JSON        Parse JSON from command-line parameter instead of from a file.
   -e|--exists PATH      Return a non-zero return status if the specified PATH does not exist. If
                         repeated each PATH must exist. Does not affect output.
   -x|--extract PATH     Extract one or more values from the result matching PATH. May be repeated.
   -X|--exclude PATH     Trim out data from being returned based on a PATH.
   -F|--fs FIELD_SEP     Sets the field separator for path (default: '/').
   -S|--no-sort          Do not sort JSON object keys (default: false).
   -d|--debug            Display debugging information on STDERR.
   -i|--indent INDENT    Indent JSON formatted output with spaces (default: 4).

PATHS

The JSON data can be filtered based on index, key matches, ranges, etc. The
JSON object is mapped to a directory-like structure (e.g. '/dict/dict_key',
'/array/0') syntax with some extra tricks. The field separator between path
parts can be changed with the -F|--fs option.

Arrays:
    By Index:
     - 'foo/0', 'foo/2', 'foo/-1' (last item)
    By Range:
     - 'foo/:' or 'foo/*' (all items within the array),
     - 'foo/2:', 'foo/:2', 'foo/1:5', 'foo/-2:' (last 2),
     - 'foo/:-2' (all but last two),
     - 'foo/1:-3' (between first and up until 3rd to last)
Dictionaries:
    Regular Expressions:
     - 'foo/b..?r' = foo/bar, foo/beer
     - 'foo/bar/.*[pP]assw(or)?d' == anything within foo/bar that looks like a password
General:
    - the root slash on paths is optional (e.g. 'foo/bar' == '/foo/bar')

EXAMPLES:

> json='{"id": 3, "name: "bob", "lols": {"a": 1, "b": 2}}'
> echo "$json" | json -x id -x lols -x lols/b -i 0
3
{"a": 1, "b": 2}
2
> echo "$json" | json -p -x lols/\*
lols_a=3
lols_b=3
""")


def get_opts(defaults, argv):
    i = 1
    opts = defaults
    while i < len(argv):
        arg = argv[i]
        if arg == '-h' or arg == '--help':
            usage()
            exit()
        elif arg == '-q' or arg == '--quiet':
            opts['quiet'] = True
        elif arg == '-p' or arg == '--pairs':
            opts['pairs'] = True
        elif arg == '-e' or arg == '--exists':
            i += 1
            if i == len(argv):
                raise Exception("Missing path to --exists.")
            opts['exists'].append(argv[i])
        elif arg == '-S' or arg == '--no-sort':
            opts['sort_keys'] = False
        elif arg == '-d' or arg == '--debug':
            opts['debug'] = True
        elif arg == '-i' or arg == '--indent':
            i += 1
            if i == len(argv):
                raise Exception("Missing number of spaces to use for --indent.")
            opts['indent'] = int(argv[i])
            if not opts['indent']:
                opts['indent'] = None
        elif arg == '-F' or arg == '--fs':
            i += 1
            if i == len(argv):
                raise Exception("Missing field separator argument to --fs.")
            if len(argv[i]) != 1:
                raise Exception("The field separator must be a single character.")
            opts['separator'] = argv[i]
        elif arg == '-j' or arg == '--json':
            i += 1
            if i == len(argv):
                raise Exception("Missing JSON string for '--json' parameter.")
            if opts['json']:
                raise Exception("JSON data cannot be provided more than once via the '--json' parameter.")
            if opts['json_file']:
                raise Exception("Unable to use both --json and specify a JSON input file.")
            opts['json'] = argv[i]
        elif arg == '-X' or arg == '--exclude':
            i += 1
            if i == len(argv):
                raise Exception("Missing path to --exclude.")
            opts['exclude'].append(argv[i])
        elif arg == '-x' or arg == '--extract':
            i += 1
            if i == len(argv):
                raise Exception("Missing path to --extract.")
            opts['extract'].append(argv[i])
        else:
            # json encoded file
            if opts['json_file']:
                raise Exception("Cannot use '%s'; the JSON input file '%s was already specified." % (arg, opts['json_file']))
            if opts['json']:
                raise Exception("Unable to use both --json and specify a JSON input file.")
            opts['json_file'] = arg
        i += 1
    return opts


def dump_obj(obj, max_len=48):
    txt = json.dumps(obj, ensure_ascii=True)
    if max_len and len(txt) > max_len:
        json_len = len(txt)
        if json_len > max_len:
            third = (max_len / 3)
            third += json_len % 3
            txt = ''.join((
                txt[0:max_len - third],
                '...',
                txt[json_len - third:]
            ))
        return txt


def parse_keys(obj, path, quiet=False):
    """
    Return the keys that we which to decend into based on the path.
    """
    keys = []
    try:
        objlen = len(obj)
    except:
        if not quiet:
            raise Exception("Unable to search for path '%s' in non-array object." % (path))
    if isinstance(obj, list):
        # treat * as a full range
        if path == '*':
            path = ':'
        # is it an index value or a range?
        path_split = path.find(':')
        if path_split == -1:
            # index, e.g. -1, 5
            path = int(path)
            # is it in range, or do we not even care?
            if not quiet and (path >= objlen or abs(path) > objlen):
                raise Exception("Invalid index %s in array." % (path))
            else:
                # otherwise return the requested element (rounding if needed on quiet mode)
                if path < 0:
                    keys = [max(path, -1 * objlen)]
                else:
                    keys = [min(path, objlen)]
        else:
            path = path.strip()
            parts = path.split(':')
            if len(parts) != 2:
                raise Exception("Invalid path part: %s." % (path))
            # catch empty; coerce to int
            if parts[0] == '':
                parts[0] = 0
            if parts[1] == '':
                parts[1] = objlen - 1
            parts = [int(part) for part in parts]
            if parts[0] < 0:
                # last X number
                parts[0] = objlen - abs(parts[0])
            if parts[1] < 0:
                # all except last X number
                parts[1] = objlen - abs(parts[1])
            elif parts[1] >= 0:
                # always need one more on positive 2nd numbers
                parts[1] = parts[1] + 1
            keys = range(parts[0], parts[1])
    elif isinstance(obj, dict):
        # write our path into a regex to compare against the keys
        try:
            path_re = re.compile(path)
        except Exception as e:
            raise Exception('Unable to compile path part "%s" to regex: %s' % (path, e))
        keys = [key for key in obj.keys() if re.match(path_re, key)]
    if not len(keys) and not quiet:
        raise Exception("Path '%s' in object '%s' not found." % (
            path,
            dump_obj(obj)
        ))
    return keys


def print_obj(obj):
    json.dumps(
        obj,
        indent=opts['indent'],
        sort_keys=opts['sort_keys'],
        ensure_ascii=True
    )


def exclude_path(obj, path, separator='/', prefix='', quiet=False, debug=False):
    """
    Exclude values from an object based on a path.
    """
    # break the path into parts
    path_parts = path.split(separator)
    # keep track of the values we've collected
    excluded = []
    if prefix:
        prefix = prefix + separator
    for key in parse_keys(obj, path_parts[0], quiet):
        # try to get the value
        try:
            value = obj[key]
        except:
            if not quiet:
                raise Exception("Invalid key '%s' not found in object '%s'." % (key, dump_obj(obj)))
            continue
        # our current path thus far
        subpath = prefix + str(key)
        # at the end of our path? we've got what we want here
        if len(path_parts) == 1:
            del obj[key]
        else:
            exclude_path(
                value,
                separator.join(path_parts[1:]),
                separator,
                subpath,
                quiet
            )


def extract_path(obj, path, separator='/', prefix='', quiet=False, debug=False):
    """
    Extract values from an object based on a path.
    """
    # break the path into parts
    path_parts = path.split(separator)
    # keep track of the values we've collected
    extracted = []
    if prefix:
        prefix = prefix + separator
    for key in parse_keys(obj, path_parts[0], quiet):
        # try to get the value
        try:
            value = obj[key]
        except:
            if not quiet:
                raise Exception("Invalid key '%s' not found in object '%s'." % (key, dump_obj(obj)))
            continue
        # our current path thus far
        subpath = prefix + str(key)
        # at the end of our path? we've got what we want here
        if len(path_parts) == 1:
            extracted.append((subpath, key, value))
        else:
            extracted += extract_path(
                value,
                separator.join(path_parts[1:]),
                separator,
                subpath,
                quiet
            )
    return extracted


def extract(
        data, indent=4, pairs=False, sort_keys=True, debug=False, quiet=False,
        separator='/', extract=None, exclude=None, exists=None, raw=False,
    ):
    if isinstance(data, str):
        obj = json.JSONDecoder().decode(data)
    else:
        obj = data
    if exists:
        for path in exists:
            try:
                extract_path(
                    obj,
                    path.strip('/'),
                    separator=separator,
                    quiet=True,
                    debug=debug
                )
            except:
                # QQ, not found
                retval = 1
                break
    # trim out any requested data
    if exclude:
        for path in exclude:
            exclude_path(
                obj,
                path.strip('/'),
                separator=separator,
                quiet=quiet,
                debug=debug
            )
    # we'll print back the obj by default
    results = [obj if raw else json.dumps(
        obj,
        ensure_ascii=True,
        sort_keys=sort_keys,
        indent=indent
    )]
    if extract:
        results = []
        name_re = re.compile(r'\W+')
        for path in extract:
            data = extract_path(
                obj,
                path.strip('/'),
                separator=separator,
                quiet=quiet,
                debug=debug
            )
            for (path, key, value) in data:
                if pairs:
                    results.append("%s=%s" % (
                        re.sub(name_re, '_', path),
                        json.dumps(
                            value,
                            ensure_ascii=True,
                            sort_keys=sort_keys,
                            indent=indent
                        )
                    ))
                else:
                    results.append(value if raw else json.dumps(
                        value,
                        ensure_ascii=True,
                        sort_keys=sort_keys,
                        indent=indent
                    ))
    return results

# stand-alone script mode
if __name__ == '__main__':
    retval = 0
    try:
        opts = get_opts({
            'indent': 4,
            'pairs': False,
            'sort_keys': True,
            'json': False,
            'debug': False,
            'quiet': False,
            'separator': '/',
            'json_file': None,
            'extract': [],
            'exclude': [],
            'exists': []
        }, sys.argv)

        # we got something, right?
        if opts['json']:
            json_data = opts['json']
        elif opts['json_file']:
            json_file = open(opts['json_file'])
            json_data = ''.join(json_file.readlines())
        else:
            # try reading stdin for data
            json_data = ''.join(sys.stdin.readlines())
        if not json_data:
            raise Exception("No JSON given to parse.")
        # we'll pass the JSON explicitly
        del opts['json']
        del opts['json_file']
        results = extract(json_data, **opts)
        for result in results:
            print(result)
    except Exception as e:
        sys.stderr.write(e.message + "\n")
        retval = 1
    sys.exit(retval)
